/*********************************************************************/
/* neblokeys.c (v. 0.2)                                              */
/* Asymmetric login key generator for neblo			     */
/* Copyright (C) 2006 - Paolo Gianrossi <paolino@member.fsf.org>     */
/* 								     */
/* This program is free software; you can redistribute it and/or     */
/* modify it under the terms of the GNU General Public License as    */
/* published by the Free Software Foundation; either version 2 of    */
/* the License, or (at your option) any later version.		     */
/*								     */
/* This program is distributed in the hope that it will be useful,   */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of    */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU */
/* General Public License for more details.			     */
/*								     */
/* You should have received a copy of the GNU General Public License */
/* along with this program; if not, write to the Free Software	     */
/* Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA	     */
/* 02111-1307, USA.						     */
/*********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

#include <neblo_secret.h>

#include <config.h>
#include <openssl/rsa.h>
#include <openssl/rand.h>

#include <time.h>
#include <getopt.h>

const char *program_version =
"neblokeys (" PACKAGE_STRING ")";
const char *program_bug_address =PACKAGE_BUGREPORT;

/* Program documentation. */
static char doc[] =
"neblokeys -- Make and test neblo asymmetric keys";

/* A description of the arguments we accept. */
static char args_doc[] = "";

/* Used by `main' to communicate with `parse_opt'. */
struct arguments
{
  char *prefix;                /* ARG1 & ARG2 */
  int verbose;
  int stout;
  int bits;
  int test;
  char *message;
  int exp;
};

#define VERBOSE_PRINT(flag, ...) if(flag){fprintf(stderr, __VA_ARGS__);}

int main (int argc, char **argv)
{
  struct arguments arguments;
  char *pubfile;
  char *secfile;
  register int retry=3;
  int v=0;
  FILE *pf, *sf;
  RSA *private=RSA_new();
  RSA *public=RSA_new();
/* The options we understand. */
/* static struct option options[] = { */
/*   {"test", 't', "MESSAGE", OPTION_ARG_OPTIONAL, "Instead of generating keys, test existing keys. " */
/*                                                 "MESSAGE will be the string used for the test."}, */
/*   {"verbose",  'v', 0,      0,  "Produce verbose output (to stderr)" }, */
/*   {"stdout",   's', 0,      0,  "Output to standard output instead of files"}, */
/*   {"bits",     'b', "LENGTH",      0,  "Make a key LENGHT bits long"}, */
/*   {"exponent", 'e', "EXP", 0, "Set key exponent to EXP"}, */
/*   {"file",   'f', "PREFIX", 0, */
/*    "Output to PREFIX.sk and PREFIX.pk instead of default files (neblo.[sk|pk])," */
/*    " or (with the `test' option), test keys PREFIX.sk and PREFIX.pk"}, */
/*   { 0 } */
/* }; */
  static struct option long_options[]=
    {
      {"version", no_argument, 0, 'V'},
      {"verbose", no_argument, 0, 'v'},
      {"test", optional_argument, 0, 't'},
      {"stdout", no_argument, 0, 's'},
      {"bits", required_argument, 0, 'b'},
      {"exponent", required_argument, 0, 'e'},
      {"file", required_argument, 0, 'f'},
      {0, 0, 0, 0}
    };

  /* Default values. */
  arguments.stout = 0;
  arguments.verbose = 0;
  arguments.prefix = "neblo";
  arguments.bits=DEFAULTBITKEYSIZE;
  arguments.exp=DEFAULTKEYEXP;
  arguments.message="This is the default test message.";
  arguments.test=0;

  while(1)
    {
      int c;
      int option_index = 0;
      c=getopt_long(argc, argv, "Vvt::b:s:e:f:", long_options, &option_index);
      if(c==-1)
	break;
      switch(c)
	{
	case 'V':
	  fprintf(stderr, "%s\nCopyright (C) 2005 - 2006 Paolo Gianrossi, Giuseppe Ciaccio\n"
		  "neblokeys comes with ABSOLUTELY NO WARRANTY.\n"
		  "You may redistribute copies of neblokeys\n"
		  "under the terms of the GNU General Public License.\n"
		  "For more information about these matters, see the file named COPYING.\n\n", program_version);
	  exit(EXIT_SUCCESS);
	  break;

	case 'v':
	  arguments.verbose=1;
	  break;
	case 't':
	  arguments.test=1;
	  if(optarg)
	    {
	      arguments.message=(char *)malloc(strlen(optarg) * sizeof(char));
	      if(!arguments.message)
		{
		  perror("parsing arguments (message)");
		  exit(EXIT_FAILURE);
		}
	      strcpy(arguments.message, optarg);
	    }
	  break;
	case 'b':
	  if(optarg)
	    arguments.bits=atol(optarg);
	  break;
	case 'e':
	  if(optarg)
	    arguments.exp=atol(optarg);
	  break;
	case 'f':
	  if(optarg)
	    {
	      arguments.prefix=(char *)malloc(strlen(optarg) * sizeof(char));
	      if(!arguments.prefix)
		{
		  perror("parsing arguments (prefix)");
		  exit(EXIT_FAILURE);
		}
	      strcpy(arguments.prefix, optarg);
	    }
	  break;
	default:
	  exit(EXIT_FAILURE);
	  break;
	}
    }

  v=arguments.verbose;

  if(!arguments.test){
    if(arguments.stout){
      VERBOSE_PRINT(v, "Keys will be output to standard output.\n");
      pf=stdout;
      sf=pf;
    }
    else
      {
	pubfile=(char *)malloc((strlen(arguments.prefix) + 5) * sizeof(char));
	secfile=(char *)malloc((strlen(arguments.prefix) + 5) * sizeof(char));
	strcpy(pubfile, arguments.prefix);
	strcat(pubfile, ".pk");
	strcpy(secfile, arguments.prefix);
	strcat(secfile, ".sk");
	VERBOSE_PRINT(v, "Public key and secret key will be output respectively to files: `%s' and `%s'\n",
		      pubfile, secfile);
      }

    srand(time(NULL));
    while (retry > 0)
      {
	VERBOSE_PRINT(v, "Generating key pair"
		      " of %d bits. Exp is %d... ", arguments.bits, arguments.exp);

	private=RSA_generate_key(arguments.bits, arguments.exp, NULL, NULL);

	VERBOSE_PRINT(v, "done.\nChecking key validity... ");

	if(!RSA_check_key(private)){
	  if(retry >1){
	    VERBOSE_PRINT(v, "FAILED! Not saving keys! Retrying...\n");
	    retry--;
	    continue;
	  }
	  else {
	    fprintf(stderr, "%s: Fatal error: Keys generated were invalid for %d consecutive times...\n"
		    "Aborting.\n", program_version, retry);
	    exit(EXIT_FAILURE);
	  }
	}
	else{
	  retry = 0;
	}
      }

    if(!arguments.stout){
      VERBOSE_PRINT(v, "done.\nSaving private key to %s... ", secfile);
      sf=fopen(secfile, "w");
      if(!sf){
	perror("Opening secret file");
	exit(EXIT_FAILURE);
      }
    } else {
      VERBOSE_PRINT(v, "done.\nWriting private key to standard output...");
    }
    /* BIGNUM *n;              // public modulus */
    /* BIGNUM *e;              // public exponent */
    /* BIGNUM *d;              // private exponent */
    /* BIGNUM *p;              // secret prime factor */
    /* BIGNUM *q;              // secret prime factor */
    /* BIGNUM *dmp1;           // d mod (p-1) */
    /* BIGNUM *dmq1;           // d mod (q-1) */
    /* BIGNUM *iqmp;           // q^-1 mod p */
    if(arguments.stout)
      fprintf(sf, "\n---Private Key Begin--\n");

    fprintf(sf, "%s\n%s\n%s\n%s\n%s\n%s\n%s\n%s",
	    BN_bn2hex(private->n),
	    BN_bn2hex(private->e),
	    BN_bn2hex(private->d),
	    BN_bn2hex(private->p),
	    BN_bn2hex(private->q),
	    BN_bn2hex(private->dmp1),
	    BN_bn2hex(private->dmq1),
	    BN_bn2hex(private->iqmp));
    if(arguments.stout)
      fprintf(sf, "\n---Private Key End--\n");

    if(!arguments.stout)
      fclose(sf);
    VERBOSE_PRINT(v, "done\n");

    if(!arguments.stout){
      VERBOSE_PRINT(v, "Saving public key to %s... ", pubfile);
      pf=fopen(pubfile, "w");
      if(!pf){
	perror("Opening public file");
	exit(EXIT_FAILURE);
      }
    } else {
      VERBOSE_PRINT(v, "Writing public key to standard output...");
    }
    public->n=private->n;
    public->e=private->e;
    if(arguments.stout)
      fprintf(pf, "\n---Public Key Begin--\n");
    fprintf(pf, "%s\n%s\n",BN_bn2hex(public->n),BN_bn2hex(public->e));
    if(arguments.stout)
      fprintf(pf, "\n---Public Key End--\n");

    if(!arguments.stout){
      fclose(pf);
      fprintf(stderr, "Done\n--Public Key Begin--\n");
      RSA_print_fp(stdout, public, 0);
      fprintf(stderr, "--Public Key End--\n");
    }
    free(pubfile);
    free(secfile);
    return 0;
  }
  else
  /* TESTING */
    {
        unsigned char *cipher;
	char *clearMsg;
	int long_msg;

	char fieldStr[MAXBYTEKEYSIZE-1];
	pubfile=(char*)malloc((strlen(arguments.prefix)+5)*sizeof(char));
	secfile=(char*)malloc((strlen(arguments.prefix)+5)*sizeof(char));
	strcpy(pubfile,arguments.prefix);
	strcat(pubfile,".pk");
	strcpy(secfile,arguments.prefix);
	strcat(secfile,".sk");
	VERBOSE_PRINT("Testing keypair %s - %s\n", pubfile, secfile);
	pf=fopen(pubfile, "r");	/* public key file */
	if(!pf){
	  perror("reading the public file");
	  exit(EXIT_FAILURE);
	}
	sf=fopen(secfile, "r");	/* secret key file */
	if(!sf){
	  perror("reading the secret file");
	  exit(EXIT_FAILURE);
	}

	_get_public_key_(pf, public, fieldStr);

	if(arguments.verbose){
	  fprintf(stderr, "--Public Key Begin--\n");
	  RSA_print_fp(stderr, public, 0);
	  fprintf(stderr, "--Public Key End--\n\n");
	}

	_get_private_key_(sf, private, fieldStr);

	/* Encriptamos el msg */
	VERBOSE_PRINT(v,"Encoding message...");
	cipher= (unsigned char *) malloc (RSA_size(public));
	long_msg= (strlen(arguments.message) +1)*sizeof(char);
	RSA_public_encrypt(long_msg, (unsigned char *)arguments.message, cipher, public, \
			   RSA_PKCS1_OAEP_PADDING);
	VERBOSE_PRINT(v,"done.\nDecoding message...");
	/* decrypt msg */
	clearMsg= (char *) malloc(RSA_size(private));
	RSA_private_decrypt(RSA_size(private), cipher, (unsigned char*)clearMsg,private, \
			    RSA_PKCS1_OAEP_PADDING);
	VERBOSE_PRINT(v, "done.\n");

	VERBOSE_PRINT(v, "The original message: `%s'\n", arguments.message);
	VERBOSE_PRINT(v, "After encoding and decoding back: `%s'\n", clearMsg);
	if(!strcmp(arguments.message, clearMsg)){
	  fprintf(stdout, "Test succeded. Keypair [%s - %s] is working.\n", pubfile, secfile);
	}
	else {
	  fprintf(stdout, "Failure! Keypair [%s - %s] is broken.\n", pubfile, secfile);
	}

	RSA_free(public);
	RSA_free(private);
	free(cipher);
	free(clearMsg);
	free(pubfile);
	free(secfile);


	return 0;


    }
  /* should never be able to get here */
  return 1;
    }

